/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#include "concentrator.hh"
#include <time.h>
#include <cstring>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <stdexcept>
#include <unistd.h>
#include <signal.h>
#include <poll.h>
#include <sys/socket.h>
#include <errno.h>

#ifdef __MACH__
#include <mach/clock.h>
#include <mach/mach.h>
#endif /* __MACH__ */

static inline int
rtclock( timespec* Now )
{
    int retval;
#ifdef __MACH__ // OS X does not have clock_gettime, use clock_get_time
    {
	int rc;
	clock_serv_t cclock;
	mach_timespec_t mts;

	host_get_clock_service(mach_host_self(), CALENDAR_CLOCK, &cclock);
	retval = clock_get_time(cclock, &mts);
	mach_port_deallocate(mach_task_self(), cclock);
	Now->tv_sec = mts.tv_sec;
        Now->tv_nsec = mts.tv_nsec;
    }
#else /* __MACH__ */
    retval = clock_gettime(CLOCK_REALTIME, Now);
#endif /* __MACH__ */
    return retval;
}

//======================================  Fix header files not containing
#ifndef POLLRDHUP
#define POLLRDHUP 0x2000
#endif

using namespace sends;
using namespace std;

//======================================  Construct template stage
concentrator::hold::hold(transaction* x, const timespec& t)
    : timeout(t), trans(x), collected(false)
{
    if (x) socket = x->ref_socket().socket_id();
    else   socket = -1;
}

//======================================  Construct template stage
concentrator::concentrator(int nThread, int nTrans)
    : seda_stage(nThread, nTrans), mTimeout(20.0), collector(false)
{
}

//======================================  Destroy template stage
concentrator::~concentrator(void) {
}

//======================================  Return stage name.
const char*
concentrator::stage_name(void) const {
    return "concentrator";
}

//======================================  Timespec compare method
bool
operator<(const struct timespec& a, const struct timespec& b) {
    if (a.tv_sec < b.tv_sec) return true;
    if (a.tv_sec > b.tv_sec) return false;
    return (a.tv_nsec < b.tv_nsec);
}

//======================================  Timespec difference
bool
operator-(const struct timespec& a, const struct timespec& b) {
    return double(long(a.tv_sec) - long(b.tv_sec)) 
	   + 1e-9 * double(a.tv_nsec - b.tv_nsec);
}

//======================================  Timespec difference
static const long k_1B = 1000000000;

timespec&
operator-=(timespec& a, const timespec& b) {
    if (a.tv_sec < b.tv_sec) {
	a.tv_sec  = 0;
	a.tv_nsec = 0;
    } else {
	a.tv_sec -= b.tv_sec;
	if (a.tv_nsec >= b.tv_nsec) {
	    a.tv_nsec -= b.tv_nsec;
	} else if (a.tv_sec) {
	    a.tv_sec--;
	    a.tv_nsec += k_1B - b.tv_nsec;
	} else {
	    a.tv_nsec = 0;
	}
    }
    return a;
}

//======================================  Assign to a timeval.
timeval&
assign_timeval(timeval& a, const timespec& b) {
    a.tv_sec  = b.tv_sec;
    a.tv_usec = b.tv_nsec / 1000;
    return a;
}

//======================================  Add double to a timespec
struct timespec& 
operator+=(struct timespec& t, double b) {
    long dt_s = long(b);
    long dt_n = long((b-dt_s)*1e9 + 0.5);
    t.tv_sec  += dt_s;
    t.tv_nsec += dt_n;
    if (t.tv_nsec >= k_1B) {
	t.tv_nsec -= k_1B;
	t.tv_sec++;
    }
    return t;
}

double
timespec_ticks(const timespec& t) {
    return double(t.tv_sec) + 1e-9 * t.tv_nsec;
}


//======================================  Perform stage processing
void
concentrator::add_transaction(transaction* trans, double timeout) {
    //------------------------------  Calculate the timeout
    timespec t;
    rtclock(&t);
    t += timeout;

    if (debug(dlvl_detail)) {
	ostringstream msg;
	msg <<" wait until " << setprecision(12) << timespec_ticks(t) << ends;
	debug_msg(dlvl_detail, trans->ident(), msg.str());
    }

    //------------------------------  Add the new timer
    thread::semlock mlock(gate);

    int N   = trans_vect.size();
    for (int i=0; i<N; i++) {
	if (trans_vect[i].socket < 0) {
	    trans_vect[i] = hold(trans, t);
	    trans = 0;
	    break;
	}
    }
    if (trans) {
	trans_vect.push_back(hold(trans, t));
    }
}

#define P_RD_FLAGS (POLLIN)
#define P_ERR_FLAGS (POLLHUP | POLLRDHUP | POLLERR | POLLNVAL)

//======================================  Perform stage processing
void
concentrator::filter(transaction* trans) {
    trans->set_stage(600);
    try {

	//------------------------------  Add the transaction to the list.
	add_transaction(trans, mTimeout);
	trans->set_stage(610);
   
    } catch(exception& e) {
	if (trans) reject(trans, string("Caught exception: ") + e.what());
    }

    //----------------------------------  Volunteer to collect
    vector<pollfd> pvec;
    while (try_collect()) {
	timespec tEnd;

	//------------------------------  collect for some ticks
	int rc = usleep(10000);

	//------------------------------  Build polling list.
	pvec.clear();
	{
	    thread::semlock mlock(gate);

	    //--------------------------  Reserve space in fd list
	    int nSock = 0;
	    int last  = 0;
	    int N = trans_vect.size();
	    for (int i=0; i<N; ++i) {
		if (trans_vect[i].socket >= 0) {
		    last = i;
		    if (!trans_vect[i].collected) nSock++;
		}
	    }
	    last++;
	    trans_vect.erase(trans_vect.begin()+last, trans_vect.end());
	    pvec.reserve(nSock);

	    //--------------------------  Get current time.
	    rtclock(&tEnd);
	    tEnd.tv_nsec = 0;
	    tEnd.tv_sec  = (tEnd.tv_sec + 2) & ~1;

	    //--------------------------  Build a list of fds
	    for (int i=0; i<N; ++i) {
		int fd = trans_vect[i].socket;
		if (fd >= 0 && !trans_vect[i].collected) {
		    pollfd pfd;
		    pfd.fd      = fd;
		    pfd.events  = P_RD_FLAGS | P_ERR_FLAGS;
		    pfd.revents = 0;
		    pvec.push_back(pfd);
		    if (trans_vect[i].timeout < tEnd) {
			tEnd = trans_vect[i].timeout;
		    }
		    trans_vect[i].collected = true;
		    trans_vect[i].trans->set_stage(620);
		}
	    }
	    release_collect();
	}
	if ( pvec.empty() ) break;

	//------------------------------  Wait for something to happen.
	if (debug(dlvl_detail)) {
	    ostringstream msg;
	    msg << "polling " << pvec.size() << " fds, timeout at: " 
		<< setprecision(12) << timespec_ticks(tEnd) << ends;
	    debug_msg(dlvl_detail, trans->ident(), msg.str());
	}
	timespec now;
	rtclock(&now);
	rc = poll(&pvec.front(), pvec.size(), int((tEnd-now)*1000.0));
	if (debug(dlvl_detail)) {
	    ostringstream msg;
	    msg << "poll rc = " << rc << ends;
	    debug_msg(dlvl_detail, trans->ident(), msg.str());
	}

	//------------------------------  Acquire control of transaction list.
	{
	    thread::semlock mlock(gate);
	    rtclock(&now);

	    //--------------------------  Scan the list, route the transaction.
	    int N = trans_vect.size();
	    int j = 0;
	    int M = pvec.size();
	    for (int i=0; i<N && j<M; ++i) {
		int fd = trans_vect[i].socket;
		if (fd == pvec[j].fd) {
		    transaction* mytrans = trans_vect[i].trans;

		    //------------------  Do something about this....
		    try {
			string disp;

			//--------------  Check for error status
			short events = pvec[j].revents;
			if ((events & P_ERR_FLAGS) != 0 ) {
			    mytrans->set_stage(640);
			    int err = 0;
			    socklen_t len = sizeof(err);
			    rc = getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, 
					    &len);
			    if (rc < 0) err = errno;
			    disp = "poll error: ";
			    if (err) {
				disp += strerror(err);
			    } else if ((events & POLLRDHUP) != 0) {
				disp += "peer hung up";
			    } else if ((events & POLLHUP) != 0) {
				disp += "connection hung up";
			    } else if ((events & POLLNVAL) != 0) {
				disp += "internal error (invalid fd)";
			    } else {
				disp += "Unreported error";
			    }
			    reject(mytrans, disp);
			    trans_vect[i].socket = -1;

			//-------------  Check input ready
			} else if ((events & P_RD_FLAGS) != 0) {
			    disp = "ready";
			    mytrans->set_stage(630);
			    put_output(mytrans);
			    trans_vect[i].socket = -1;

			//-------------  Check for timeout
			} else if (trans_vect[i].timeout < now) {
			    disp = "timed out";
			    trans_vect[i].socket = -1;
			    mytrans->set_stage(660);
			    reject(mytrans, disp);

			//--------------  Still waiting...
			} else {
			    disp = "kept";
			    mytrans->set_stage(670);
			}

			if (debug(dlvl_detail)) {
			    ostringstream msg;
			    msg << "found fd: " << fd << " " << disp << ends;
			    debug_msg(dlvl_detail, mytrans->ident(), msg.str());
			}

		    //------------------  Don't lose a transaction.
		    } catch (exception& e) {
			reject(mytrans, e.what());
			trans_vect[i].socket = -1;
		    }

		    //------------------  Free for next collector;
		    trans_vect[i].collected = false;
		    j++;
		} 
	    }
	    //--------------------------  End of list scan...
	}
    }
}

//======================================  Thread initialization.
void
concentrator::init_thread(void) {
    //sigignore(SIGPIPE);
    siginterrupt(SIGPIPE, 1);
}

//======================================  Lock transaction database.
void 
concentrator::lock(void) {
    gate.lock();
}

//======================================  Exit collection mode
void 
concentrator::release_collect(void) {
    collector = false;
}

//======================================  Exit collection mode
void 
concentrator::set_parameters(const stage_pars& p) {
    mTimeout = p._timeout;
    seda_stage::set_parameters(p);
}

//======================================  Try to enter collection mode
bool 
concentrator::try_collect(void) {
    thread::semlock mlock(gate);
    bool haveit = !collector;
    collector = true;
    return haveit;
}

//======================================  Unlock transaction database.
void 
concentrator::unlock(void) {
    gate.unlock();
}
